package org.fortiss.tooling.kernel.internal;

import static org.eclipse.emf.ecore.xmi.XMLResource.OPTION_PROCESS_DANGLING_HREF;
import static org.eclipse.emf.ecore.xmi.XMLResource.OPTION_PROCESS_DANGLING_HREF_DISCARD;

import java.io.ByteArrayOutputStream;
import java.io.FileOutputStream;
import java.io.ObjectOutputStream;
import java.math.BigInteger;
import java.security.MessageDigest;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.fortiss.tooling.kernel.ToolingKernelActivator;
import org.fortiss.tooling.kernel.extension.IConstraintVerifier;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.introspection.KernelIntrospectionSystemService;
import org.fortiss.tooling.kernel.model.constraints.ConstrainedWithChecksum;
import org.fortiss.tooling.kernel.model.constraints.ConstraintsFactory;
import org.fortiss.tooling.kernel.model.constraints.IConstrained;
import org.fortiss.tooling.kernel.model.constraints.IConstraint;
import org.fortiss.tooling.kernel.model.constraints.IConstraintVerificationStatus;
import org.fortiss.tooling.kernel.service.IConstraintVerificationService;
import org.fortiss.tooling.kernel.service.IPersistencyService;
import org.fortiss.tooling.kernel.service.base.EObjectAwareServiceBase;
import org.fortiss.tooling.kernel.utils.EcoreUtils;
import org.fortiss.tooling.kernel.utils.LoggingUtils;

/**
 * Implementation of {@link IConstraintVerificationService}.
 * 
 * @author aravantinos
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 90C7294160827C1BA216A969E3B21981
 */
public final class ConstraintVerificationService extends
		EObjectAwareServiceBase<IConstraintVerifier<IConstraint>> implements
		IConstraintVerificationService {

	/** Constructor. */
	public ConstraintVerificationService() {
		KernelIntrospectionSystemService.INSTANCE.registerService(this);
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionDescription() {
		// TODO (FH): implement
		return "";
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		// TODO (FH): check and correct if necessary
		return "Constraint Verification Service";
	}

	/** {@inheritDoc} */
	@Override
	public void verify(IConstraint constraint) {
		IConstraintVerifier<IConstraint> verifier = getFirstVerifier(constraint);
		if(verifier != null) {
			ITopLevelElement modelContext =
					IPersistencyService.INSTANCE.getTopLevelElementFor(constraint);
			// It can happen that <code>modelContext</code> is null, e.g., if the verification is
			// triggered while the constraint is actually being removed.
			if(modelContext == null) {
				return;
			}
			modelContext.runAsCommand(new Runnable() {
				@Override
				public void run() {
					verifier.verify(constraint);
					// Updating checksums must be done *after* verification because the verification
					// can have side effects.
					updateChecksums(constraint);
					constraint.getVerificationStatus().setConstraint(constraint);
				}
			});
		}
	}

	/** {@inheritDoc} */
	@Override
	public List<IFix> fixes(IConstraintVerificationStatus status) {
		List<IFix> fixes = null;
		if(status != null) {
			IConstraintVerifier<IConstraint> verifier = getFirstVerifier(status.getConstraint());
			if(verifier != null) {
				fixes = verifier.fixes(status);
			}
		}
		return fixes != null ? fixes : new ArrayList<IFix>();
	}

	/**
	 * @param constraint
	 * @return <code>true</code> if <code>constraint</code> is up to date, <code>false</code>
	 *         otherwise.
	 */
	@Override
	public boolean isUpToDate(IConstraint constraint) {
		for(ConstrainedWithChecksum cwc : constraint.getConstrainedsWithChecksum()) {
			BigInteger computeCheckSum = computeCheckSum(cwc.getConstrained(), constraint);
			if(computeCheckSum == null || !computeCheckSum.equals(cwc.getChecksum())) {
				return false;
			}
		}
		return true;
	}

	/**
	 * @param constraint
	 *            Updates all the checksums of <code>constraint</code>.
	 */
	private void updateChecksums(IConstraint constraint) {
		List<ConstrainedWithChecksum> newCwcs = new ArrayList<ConstrainedWithChecksum>();
		for(ConstrainedWithChecksum cwc : constraint.getConstrainedsWithChecksum()) {
			ConstrainedWithChecksum newCwc =
					ConstraintsFactory.eINSTANCE.createConstrainedWithChecksum();
			newCwc.setConstrained(cwc.getConstrained());
			newCwc.setChecksum(computeCheckSum(cwc.getConstrained(), constraint));
			newCwcs.add(newCwc);
		}
		EcoreUtils.disableNotificationsOf(constraint);
		constraint.getConstrainedsWithChecksum().clear();
		constraint.getConstrainedsWithChecksum().addAll(newCwcs);
		EcoreUtils.enableNotificationsOf(constraint);
	}

	/**
	 * Flag to print the object which is actually checksummed in a file.
	 * TODO: remove when the experimental phase for constraints is over.
	 */
	private final boolean DEBUG = false;

	/**
	 * @param constrained
	 * @param constraint
	 * @return the checksum of <code>constrained</code>. Note that <code>constraint</code> is
	 *         necessary because it has a potential impact on what is relevant for the checksum or
	 *         not.
	 */
	protected BigInteger computeCheckSum(IConstrained constrained, IConstraint constraint) {
		if(constrained == null) {
			return null;
		}
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		final ObjectOutputStream oos;
		final EObject hashRelevant = getChecksumRelevantEObject(constrained, constraint);
		try {
			final ResourceSet rset = new ResourceSetImpl();
			oos = new ObjectOutputStream(baos);
			Resource res = rset.createResource(URI.createURI("temp"));
			res.getContents().add(hashRelevant);
			Map<String, String> options = new HashMap<>();

			// The option below is to allow computing the hashsum of an incomplete ecore sub-model,
			// which is wanted in case parts of the model which are not checksum relevant are still
			// referred to.
			options.put(OPTION_PROCESS_DANGLING_HREF, OPTION_PROCESS_DANGLING_HREF_DISCARD);

			res.save(oos, options);
			if(DEBUG) {
				FileOutputStream file = new FileOutputStream("/home/aravantinos/temp.xml");
				ObjectOutputStream oos2 = new ObjectOutputStream(file);
				res.save(oos2, options);
				oos2.close();
			}
			oos.close();
			MessageDigest m = MessageDigest.getInstance("SHA1");
			m.update(baos.toByteArray());
			res.getContents().remove(hashRelevant);
			res.delete(options);
			return new BigInteger(1, m.digest());
		} catch(Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * @param constrained
	 * @param constraint
	 * @return The object to be used to compute the check sum.
	 *         Generally, all verification statuses and checksums are removed. Each verifier can
	 *         also provide per-constraint adaptations.
	 */
	protected EObject getChecksumRelevantEObject(IConstrained constrained, IConstraint constraint) {
		EObject res = EcoreUtils.copy(constrained);
		IConstraintVerifier<IConstraint> verifier = getFirstVerifier(constraint);
		if(verifier != null) {
			verifier.getChecksumRelevantEObject(res);
		}
		return res;
	}

	/**
	 * @param constraint
	 * @return the first registered verifier which can verify <code>constraint</code>
	 */
	protected IConstraintVerifier<IConstraint> getFirstVerifier(IConstraint constraint) {
		List<IConstraintVerifier<IConstraint>> list = getRegisteredHandlers(constraint.getClass());
		if(list == null || list.isEmpty()) {
			LoggingUtils.error(ToolingKernelActivator.getDefault(),
					String.format("Cannot find verifier for constraint %s", constraint.getClass()));
			return null;
		}
		// get(0) because we know the list is not empty and because we want the first element
		return list.get(0);
	}

	/** {@inheritDoc} */
	@Override
	protected String getExtensionPointName() {
		return "org.fortiss.tooling.kernel.constraintVerifier";
	}

	/** {@inheritDoc} */
	@Override
	protected String getConfigurationElementName() {
		return "constraintVerifier";
	}

	/** {@inheritDoc} */
	@Override
	protected String getHandlerClassAttribute() {
		return "constraintVerifier";
	}

	/** {@inheritDoc} */
	@Override
	protected String getClassAttribute() {
		return "constraint";
	}

	/** {@inheritDoc} */
	@Override
	protected String getClassConfigurationElement() {
		return "constraint";
	}

	/** {@inheritDoc} */
	@Override
	public void setConstrained(IConstraint constraint, IConstrained constrained, int index) {
		EList<ConstrainedWithChecksum> cwcs = constraint.getConstrainedsWithChecksum();
		if(cwcs.size() <= index) {
			for(int i = cwcs.size(); i <= index; i++) {
				cwcs.add(i, ConstraintsFactory.eINSTANCE.createConstrainedWithChecksum());
			}
		}
		if(!constrained.getConstraints().contains(constraint)) {
			constrained.getConstraints().add(constraint);
		}
		ConstrainedWithChecksum cwc = cwcs.get(index);
		cwc.setConstrained(constrained);
		cwc.setChecksum(computeCheckSum(constrained, constraint));
	}
}
